##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
'''
Functions for refreshing products.
$Id: RefreshFuncs.py,v 1.4 2001/02/12 20:26:45 shane Exp $
'''

import os, sys
from time import time
from string import split, join
import InitProduct
import Products
from ExtensionClass import Base
from Globals import PersistentMapping
from zLOG import LOG, ERROR, INFO

global_classes_timestamp = 0
products_mod_times = {}

_marker = []  # create a new marker object.

# Functions for storing and retrieving the auto-refresh state for
# each product.

def _getCentralRefreshData(jar, create=0):
    root = jar.root()
    if root.has_key('RefreshData'):
        rd = root['RefreshData']
    else:
        rd = PersistentMapping()
        if create:
            root['RefreshData'] = rd
    return rd
    
def isAutoRefreshEnabled(jar, productid):
    rd = _getCentralRefreshData(jar)
    ids = rd.get('auto', None)
    if ids:
        return ids.get(productid, 0)
    else:
        return 0

def enableAutoRefresh(jar, productid, enable):
    productid = str(productid)
    rd = _getCentralRefreshData(jar, 1)
    ids = rd.get('auto', None)
    if ids is None:
        if enable:
            rd['auto'] = ids = PersistentMapping()
        else:
            return
    if enable:
        ids[productid] = 1
    else:
        if ids.has_key(productid):
            del ids[productid]

def listAutoRefreshableProducts(jar):
    rd = _getCentralRefreshData(jar)
    auto = rd.get('auto', None)
    if auto:
        ids = []
        for k, v in auto.items():
            if v:
                ids.append(k)
        return ids
    else:
        return ()

def getDependentProducts(jar, productid):
    rd = _getCentralRefreshData(jar)
    products = rd.get('products', None)
    if products is None:
        return ()
    product = products.get(productid, None)
    if product is None:
        return ()
    return product.get('dependent_products', ())

def setDependentProducts(jar, productid, dep_ids):
    productid = str(productid)
    rd = _getCentralRefreshData(jar, 1)
    products = rd.get('products', None)
    if products is None:
        rd['products'] = products = PersistentMapping()
    product = products.get(productid, None)
    if product is None:
        products[productid] = product = PersistentMapping()
    product['dependent_products'] = tuple(map(str, dep_ids))


# Functions for performing refresh.

def getReloadVar(module):
    reload_var = getattr(module, '__refresh_module__', _marker)
    if reload_var is _marker:
        reload_var = getattr(module, '__reload_module__', _marker)
    if reload_var is _marker:
        reload_var = 1
    return reload_var    

def listRefreshableModules(productid):
    prefix = "Products.%s" % productid
    prefixdot = prefix + '.'
    lpdot = len(prefixdot)
    rval = []
    for name, module in sys.modules.items():
        if module and (name == prefix or name[:lpdot] == prefixdot):
            reload_var = getReloadVar(module)
            if callable(reload_var) or reload_var:
                rval.append((name, module))
    return rval

def performRefresh(jar, productid):
    '''Attempts to perform a refresh operation.  self is the Product
    object.
    '''
    setupModTimes(productid)  # Refresh again only if changed again.

    modlist = listRefreshableModules(productid)

    reload_normal = []
    reload_by_func = []

    for name, module in modlist:
        # Remove the __import_error__ attribute.
        try: del module.__import_error__
        except: pass
        # Ask the module how it should be reloaded.
        reload_var = getReloadVar(module)
        if callable(reload_var):
            reload_by_func.append(reload_var)
        else:
            reload_normal.append(module)

    for m in reload_normal:
        reload(m)
    for func in reload_by_func:
        func()

    # Reinitialize and reinstall the product.
    InitProduct.import_product(productid)
    app = jar.root()['Application']
    InitProduct.install_product(app, productid)

def performLoggedRefresh(jar, productid, raise_exc=0):
    try:
        LOG('Refresh', INFO, 'Refreshing product %s' % productid)
        performRefresh(jar, productid)
    except:
        LOG('Refresh', ERROR, 'Exception while refreshing %s'
            % productid, error=sys.exc_info())
        if raise_exc:
            raise
        return 0
    else:
        return 1

def performFullRefresh(jar, productid, raise_exc=0):
    if performLoggedRefresh(jar, productid, raise_exc):
        dep_ids = getDependentProducts(jar, productid)
        for dep_id in dep_ids:
            if isAutoRefreshEnabled(jar, dep_id):
                performLoggedRefresh(jar, dep_id, raise_exc)

# Functions for clearing ZODB caches.

def updateGlobalClassesTimestamp():
    global global_classes_timestamp
    global_classes_timestamp = time()

def mayClearCache(jar):
    # An opportunity to reset the cache.
    if getattr(jar, '_classes_timestamp', 0) != global_classes_timestamp:
        jar._classes_timestamp = global_classes_timestamp
        jar._invalidated.clear()
        # Clear the cache by re-creating it.
        origCache = jar._cache
        from ZODB.cPickleCache import PickleCache
        jar._cache = PickleCache(jar, origCache.cache_size,
                                 origCache.cache_age)

# Functions for quickly scanning the dates of product modules.

def tryFindProductDirectory(productid):
    path_join = os.path.join
    isdir = os.path.isdir
    exists = os.path.exists

    for products_dir in Products.__path__:
        product_dir = path_join(products_dir, productid)
        if not isdir(product_dir): continue
        if not exists(path_join(product_dir, '__init__.py')):
            if not exists(path_join(product_dir, '__init__.pyc')):
                continue
        return product_dir
    return None

def tryFindModuleFilename(product_dir, filename):
    # Try different variations of the filename of a module.
    path_join = os.path.join
    isdir = os.path.isdir
    exists = os.path.exists

    found = None
    fn = path_join(product_dir, filename + '.py')
    if exists(fn):
        found = fn
    if not found:
        fn = fn + 'c'
        if exists(fn):
            found = fn
    if not found:
        fn = path_join(product_dir, filename)
        if isdir(fn):
            fn = path_join(fn, '__init__.py')
            if exists(fn):
                found = fn
            else:
                fn = fn + 'c'
                if exists(fn):
                    found = fn
    return found

def setupModTimes(productid):
    mod_times = []
    product_dir = tryFindProductDirectory(productid)
    if product_dir is not None:
        modlist = listRefreshableModules(productid)

        path_join = os.path.join
        exists = os.path.exists

        for name, module in modlist:
            splitname = split(name, '.')[2:]
            if not splitname:
                filename = '__init__'
            else:
                filename = apply(path_join, splitname)
            found = tryFindModuleFilename(product_dir, filename)

            if found:
                try: mtime = os.stat(found)[8]
                except: mtime = 0
                mod_times.append((found, mtime))
    products_mod_times[productid] = mod_times

def checkModTimes(productid):
    # Returns 1 if there were changes.
    mod_times = products_mod_times.get(productid, None)
    if mod_times is None:
        # Initialize the mod times.
        setupModTimes(productid)
        return 0
    for filename, mod_time in mod_times:
        try: mtime = os.stat(filename)[8]
        except: mtime = 0
        if mtime != mod_time:
            # Something changed!
            return 1
    return 0

##def mayInitModTimes(productid):
##    # Initialize the mod times if not yet created.
##    mod_times = products_mod_times.get(productid, None)
##    if mod_times is None:
##        setupModTimes(productid)


# Functions for performing auto-refresh.

next_auto_refresh_check = 0

def checkAutoRefresh(jar):
    # Note: this function is NOT allowed to change the database!
    global next_auto_refresh_check
    now = time()
    if next_auto_refresh_check and next_auto_refresh_check > now:
        # Not enough time has passed.
        return ()
    next_auto_refresh_check = now + 2  # 2 seconds.

    rd = _getCentralRefreshData(jar)
    ids = rd.get('auto', None)
    if not ids:
        return ()
    auto_refresh_ids = []
    for productid in ids.keys():
        if checkModTimes(productid):
            auto_refresh_ids.append(productid)
    return auto_refresh_ids

def finishAutoRefresh(jar, productids):
    # This function is allowed to change the database.
    for productid in productids:
        performFullRefresh(jar, productid)

















if 0:
    # These are some interesting utility functions developed while
    # working on Refresh.  They may or may not be useful.

    class dummyClass: pass
    class dummyClass2 (Base): pass

    ClassTypes = (type(dummyClass), type(dummyClass2))

    def reloadModule(m, old_klasses):
        # Stuffs new attributes into old classes.
        # Doesn't work?
        reload(m)

        if old_klasses:
            # Shove the new attributes into the old class definitions
            # so that subclasses get updated.
            name = m.__name__
            newmod = sys.modules[name]
            for key, oldvalue in old_klasses.items():
                newvalue = newmod.__dict__.get(key, None)
                if (type(newvalue) in ClassTypes
                    and getattr(newvalue, '__module__', None) == name
                    and hasattr(newvalue, '__dict__')):
                    # Add and update attributes.
                    for k, v in newvalue.__dict__.items():
                        if not oldvalue.__dict__.has_key(k):
                            print 'adding', name, key, k
                        elif oldvalue.__dict__[k] is not v:
                            #print 'changed', name, key, k
                            pass
                        oldvalue.__dict__[k] = v
                    # Don't remove old attributes--they are often
                    # generated on the fly.


    def reassembleBases(productId):
        # Stuffs new base classes into old classes.
        # Doesn't work?
        prefix = "Products.%s" % productId
        prefixdot = prefix + '.'

        for name, module in sys.modules.items():
            if module:
                if name == prefix or name[:len(prefixdot)] == prefixdot:
                    # No need to change.
                    continue
                else:
                    for key, value in module.__dict__.items():
                        if (type(value) in ClassTypes
                            and getattr(value, '__module__', None) == name
                            and hasattr(value, '__bases__')):
                            bases = list(value.__bases__)
                            changed = 0
                            for idx in range(len(bases)):
                                base = bases[idx]
                                n = getattr(base, '__module__', '')
                                if n == prefix or n[:len(prefixdot)] == prefixdot:
                                    # Replace this base class.
                                    klass = getattr(sys.modules[n],
                                                    base.__name__, None)
                                    if klass is not None and klass != base:
                                        print 'replacing base', `n`, 'of', `key`
                                        print base, klass
                                        bases[idx] = klass
                                        changed = 1
                            if changed:
                                value.__bases__ = tuple(bases)


    def findModulesWithSubclasses(productId):
        # Finds all modules in other products that use a base class
        # from the given product.  Tested in Python 1.5.2.
        prefix = "Products.%s" % productId
        prefixdot = prefix + '.'

        found = {}

        for name, module in sys.modules.items():
            if module:
                if name == prefix or name[:len(prefixdot)] == prefixdot:
                    continue
                else:
                    for key, value in module.__dict__.items():
                        if (type(value) in ClassTypes
                            and getattr(value, '__module__', None) == name
                            and hasattr(value, '__bases__')):
                            for base in value.__bases__:
                                n = getattr(base, '__module__', '')
                                if (n == prefix or n[:len(prefixdot)]
                                    == prefixdot):
                                    found[name] = 1
        return found
